@@ -126,7 +126,19 @@ char *ht_time(pool *p, time_t t, char *fmt, int gmt) {
 }
 
 char *gm_timestr_822(pool *p, time_t sec) {
-    return ht_time(p, sec, HTTP_TIME_FORMAT, 1);
+    static const char *const days[7]=
+       {"Sun","Mon", "Tue", "Wed", "Thu", "Fri", "Sat"};
+    char ts[50];
+    struct tm *tms;
+
+    tms = gmtime(&sec);
+
+/* RFC date format; as strftime '%a, %d %b %Y %T GMT' */
+    sprintf(ts, "%s, %.2d %s %d %.2d:%.2d:%.2d GMT", days[tms->tm_wday],
+	    tms->tm_mday, month_snames[tms->tm_mon], tms->tm_year + 1900,
+	    tms->tm_hour, tms->tm_min, tms->tm_sec);
+
+    return pstrdup (p, ts);
 }
 
 /* What a pain in the ass. */
@@ -183,6 +195,29 @@ int strcmp_match(char *str, char *exp) {
     return (str[x] != '\0');
 }
 
+int strcasecmp_match(char *str, char *exp) {
+    int x,y;
+
+    for(x=0,y=0;exp[y];++y,++x) {
+        if((!str[x]) && (exp[y] != '*'))
+            return -1;
+        if(exp[y] == '*') {
+            while(exp[++y] == '*');
+            if(!exp[y])
+                return 0;
+            while(str[x]) {
+                int ret;
+                if((ret = strcasecmp_match(&str[x++],&exp[y])) != 1)
+                    return ret;
+            }
+            return -1;
+        } else 
+            if((exp[y] != '?') && (tolower(str[x]) != tolower(exp[y])))
+                return 1;
+    }
+    return (str[x] != '\0');
+}
+
 int is_matchexp(char *str) {
     register int x;
 
@@ -320,6 +355,26 @@ char *getword(pool* atrans, char **line, char stop) {
     
     return res;
 }
+char *getword_nulls(pool* atrans, char **line, char stop) {
+    int pos = ind(*line, stop);
+    char *res;
+
+    if (pos == -1) {
+        res = pstrdup (atrans, *line);
+	*line += strlen (*line);
+	return res;
+    }
+  
+    res = palloc(atrans, pos + 1);
+    strncpy (res, *line, pos);
+    res[pos] = '\0';
+    
+    ++pos;
+    
+    *line += pos;
+    
+    return res;
+}
 
 /* Get a word, (new) config-file style --- quoted strings and backslashes
  * all honored
@@ -427,6 +482,52 @@ cfg_getline(char *s, int n, FILE *f) {
     }
 }
 
+/* Retrieve a token, spacing over it and returning a pointer to
+ * the first non-white byte afterwards.  Note that these tokens
+ * are delimited by semis and commas; and can also be delimited
+ * by whitespace at the caller's option.
+ */
+
+char *get_token (pool *p, char **accept_line, int accept_white)
+{
+    char *ptr = *accept_line;
+    char *tok_start;
+    char *token;
+    int tok_len;
+  
+    /* Find first non-white byte */
+    
+    while (*ptr && isspace(*ptr))
+      ++ptr;
+
+    tok_start = ptr;
+    
+    /* find token end, skipping over quoted strings.
+     * (comments are already gone).
+     */
+    
+    while (*ptr && (accept_white || !isspace(*ptr))
+	   && *ptr != ';' && *ptr != ',')
+    {
+	if (*ptr++ == '"')
+	    while (*ptr)
+	        if (*ptr++ == '"') break;
+    }
+	  
+    tok_len = ptr - tok_start;
+    token = palloc (p, tok_len + 1);
+    strncpy (token, tok_start, tok_len);
+    token[tok_len] = '\0';
+    
+    /* Advance accept_line pointer to the next non-white byte */
+
+    while (*ptr && isspace(*ptr))
+      ++ptr;
+
+    *accept_line = ptr;
+    return token;
+}
+
 char *escape_shell_cmd(pool *p, char *s) {
     register int x,y,l;
     char *cmd;
@@ -436,7 +537,16 @@ char *escape_shell_cmd(pool *p, char *s) {
     strcpy (cmd, s);
     
     for(x=0;cmd[x];x++) {
-        if(ind("&;`'\"|*?~<>^()[]{}$\\",cmd[x]) != -1){
+    
+#ifdef __EMX__
+        /* Don't allow '&' in parameters under OS/2. */
+        /* This can be used to send commands to the shell. */
+        if (cmd[x] == '&') {
+            cmd[x] = ' ';
+        }
+#endif
+
+        if(ind("&;`'\"|*?~<>^()[]{}$\\\n",cmd[x]) != -1){
             for(y=l+1;y>x;y--)
                 cmd[y] = cmd[y-1];
             l++; /* length has been increased */
@@ -531,6 +641,9 @@ prefixed with "./" (or the ':' escaped). In the case of Unix, this means
 leaving '/' alone, but otherwise doing what escape_path_segment() does. For
 efficiency reasons, we don't use escape_path_segment(), which is provided for
 reference. Again, RFC 1808 is where this stuff is defined.
+
+If partial is set, os_escape_path() assumes that the path will be appended to
+something with a '/' in it (and thus does not prefix "./").
 */
 
 char *escape_path_segment(pool *p, const char *segment) {
@@ -552,16 +665,20 @@ char *escape_path_segment(pool *p, const char *segment) {
     return copy;
 }
 
-char *os_escape_path(pool *p,const char *path) {
+char *os_escape_path(pool *p,const char *path,int partial) {
   char *copy=palloc(p,3*strlen(path)+3);
   char *s=copy;
-  int colon=ind(path,':');
-  int slash=ind(path,'/');
 
-  if(colon >= 0 && (colon < slash || slash < 0))
+  if(!partial)
     {
-      *s++='.';
-      *s++='/';
+      int colon=ind(path,':');
+      int slash=ind(path,'/');
+
+      if(colon >= 0 && (colon < slash || slash < 0))
+	{
+	  *s++='.';
+	  *s++='/';
+	}
     }
   for( ; *path ; ++path)
     {
@@ -679,6 +796,10 @@ int is_url(char *u) {
 }
 
 int can_exec(struct stat *finfo) {
+#ifdef __EMX__
+    /* OS/2 dosen't have Users and Groups */
+    return (finfo->st_mode & S_IEXEC);
+#else    
     if(user_id == finfo->st_uid)
         if(finfo->st_mode & S_IXUSR)
             return 1;
@@ -686,6 +807,7 @@ int can_exec(struct stat *finfo) {
         if(finfo->st_mode & S_IXGRP)
             return 1;
     return (finfo->st_mode & S_IXOTH);
+#endif    
 }
 
 #ifdef NEED_STRDUP
@@ -865,16 +987,39 @@ struct in_addr get_local_addr(int sd) {
 }
 #endif
 
-unsigned long get_virthost_addr (char *w, int wild_allowed) {
+/*
+ * Parses a host of the form <address>[:port]
+ * :port is permitted if 'port' is not NULL
+ */
+unsigned long get_virthost_addr (char *w, short int *ports) {
     struct hostent *hep;
     unsigned long my_addr;
-    
-    if (wild_allowed && !strcmp(w, "*")) 
+    char *p;
+
+    p = strchr(w, ':');
+    if (ports != NULL)
+    {
+	*ports = 0;
+	if (p != NULL && strcmp(p+1, "*") != 0) *ports = atoi(p+1);
+    }
+
+    if (p != NULL) *p = '\0';
+    if (strcmp(w, "*") == 0)
+    {
+	if (p != NULL) *p = ':';
 	return htonl(INADDR_ANY);
+    }
 	
+#ifdef DGUX
+    my_addr = inet_network(w);
+#else
     my_addr = inet_addr(w);
+#endif
     if (my_addr != ((unsigned long) 0xffffffff))
+    {
+	if (p != NULL) *p = ':';
 	return my_addr;
+    }
 
     hep = gethostbyname(w);
 	    
@@ -890,82 +1035,18 @@ unsigned long get_virthost_addr (char *w, int wild_allowed) {
 	exit(1);
     }
 	    
+    if (p != NULL) *p = ':';
+
     return ((struct in_addr *)(hep->h_addr))->s_addr;
 }
 
-#ifdef DEBUG_ACCESS
-/* Host name and IP address used if getpeername() fails.
- * The most frequent cause of this is running the server in inetd-mode
- * with the tty as input and output, under the debugger.
- */
-#define NO_SUCH_HOST_NAME "volterra.ai.mit.edu"
-#define NO_SUCH_HOST_ADDR "128.52.39.173"
-#else
-#define NO_SUCH_HOST_NAME "UNKNOWN_HOST"
-#define NO_SUCH_HOST_ADDR "UNKNOWN_IP"
-#endif
-
-void get_remote_host(conn_rec *conn)
-{
-    struct sockaddr addr;
-    int len;
-    struct in_addr *iaddr;
-#if defined(MAXIMUM_DNS) || !defined(MINIMAL_DNS)
-    struct hostent *hptr;
-#endif
-
-    len = sizeof(struct sockaddr);
-
-    if ((getpeername(fileno(conn->client), &addr, &len)) < 0) {
-	conn->remote_name = conn->remote_host = NO_SUCH_HOST_NAME;
-	conn->remote_ip = NO_SUCH_HOST_ADDR;
-        return;
-    }
-
-    iaddr = &(((struct sockaddr_in *)&addr)->sin_addr);
-#ifndef MINIMAL_DNS
-    hptr = gethostbyaddr((char *)iaddr, sizeof(struct in_addr), AF_INET);
-    if(hptr) {
-        conn->remote_host = pstrdup(conn->pool, (void *)hptr->h_name);
-        conn->remote_name = conn->remote_host;
-        str_tolower (conn->remote_host);
-    }
-    else 
-#endif
-        conn->remote_host = NULL;
-
-#ifdef MAXIMUM_DNS
-    /* Grrr. Check THAT name to make sure it's really the name of the addr. */
-    /* Code from Harald Hanche-Olsen <hanche@imf.unit.no> */
-    if (conn->remote_host) {
-        char **haddr;
-
-        hptr = gethostbyname(conn->remote_host);
-        if (hptr) {
-            for(haddr=hptr->h_addr_list;*haddr;haddr++) {
-                if(((struct in_addr *)(*haddr))->s_addr == iaddr->s_addr)
-                    break;
-            }
-        }
-        if((!hptr) || (!(*haddr)))
-	    conn->remote_host = conn->remote_name = NULL;
-    }
-#endif
-    conn->remote_ip = pstrdup (conn->pool, inet_ntoa(*iaddr));
-    if(!conn->remote_host){ 	
-        conn->remote_name = conn->remote_ip;
-    }
-    if (!conn->remote_name) {
-	conn->remote_name = "UNKNOWN_HOST";
-    }
-}
 
 #ifdef NOTDEF    
     
 char *get_remote_logname(FILE *fd) {
     int len;
     char *result;
-#if defined(NEXT) || defined(BSD4_4) || defined(SOLARIS2) || defined(LINUX)
+#if defined(NEXT) || defined(BSD4_4) || defined(SOLARIS2) || defined(LINUX) || defined(__EMX__)
     struct sockaddr sa_server, sa_client;
 #else
     struct sockaddr_in sa_server,sa_client;
@@ -1076,6 +1157,32 @@ char *uudecode(pool *p, char *bufcoded) {
     return bufplain;
 }
 
+#ifdef __EMX__
+void os2pathname(char *path) {
+    char newpath[MAX_STRING_LEN];
+    int loop;
+    int offset;
+
+    offset = 0;
+    for (loop=0; loop < (strlen(path) + 1); loop++) {
+        if (path[loop] == '/') {
+            newpath[offset] = '\\';
+            /*
+            offset = offset + 1;
+            newpath[offset] = '\\';
+            */
+        } else
+            newpath[offset] = path[loop];
+        offset = offset + 1;
+    };
+    /* Debugging code */
+    /* fprintf(stderr, "%s \n", newpath); */
+
+    strcpy(path, newpath);
+};
+#endif
+
+
 #ifdef NEED_STRERROR
 char *
 strerror (int err) {
